@@ -16,7 +16,7 @@
 
 /**
  * SECTION:soup-cookie-jar
- * @short_description: Automatic cookie handling for #SoupSession
+ * @short_description: Automatic cookie handling for SoupSession
  *
  * A #SoupCookieJar stores #SoupCookie<!-- -->s and arrange for them
  * to be sent with the appropriate #SoupMessage<!-- -->s.
@@ -371,8 +371,8 @@ get_cookies (SoupCookieJar *jar, SoupURI *uri, gboolean for_http, gboolean copy_
  * almost certainly be setting @for_http to %FALSE if you are calling
  * this.
  *
- * Return value: the cookies, in string form, or %NULL if there are no
- * cookies for @uri.
+ * Return value: (nullable): the cookies, in string form, or %NULL if
+ * there are no cookies for @uri.
  *
  * Since: 2.24
  **/
@@ -655,21 +655,7 @@ process_set_cookie_header (SoupMessage *msg, gpointer user_data)
 }
 
 static void
-soup_cookie_jar_request_queued (SoupSessionFeature *feature,
-				SoupSession *session,
-				SoupMessage *msg)
-{
-	soup_message_add_header_handler (msg, "got-headers",
-					 "Set-Cookie",
-					 G_CALLBACK (process_set_cookie_header),
-					 feature);
-}
-
-static void
-soup_cookie_jar_request_started (SoupSessionFeature *feature,
-				 SoupSession *session,
-				 SoupMessage *msg,
-				 SoupSocket *socket)
+msg_starting_cb (SoupMessage *msg, gpointer feature)
 {
 	SoupCookieJar *jar = SOUP_COOKIE_JAR (feature);
 	char *cookies;
@@ -683,6 +669,21 @@ soup_cookie_jar_request_started (SoupSessionFeature *feature,
 		soup_message_headers_remove (msg->request_headers, "Cookie");
 }
 
+static void
+soup_cookie_jar_request_queued (SoupSessionFeature *feature,
+				SoupSession *session,
+				SoupMessage *msg)
+{
+	g_signal_connect (msg, "starting",
+			  G_CALLBACK (msg_starting_cb),
+			  feature);
+
+	soup_message_add_header_handler (msg, "got-headers",
+					 "Set-Cookie",
+					 G_CALLBACK (process_set_cookie_header),
+					 feature);
+}
+
 static void
 soup_cookie_jar_request_unqueued (SoupSessionFeature *feature,
 				  SoupSession *session,
@@ -696,7 +697,6 @@ soup_cookie_jar_session_feature_init (SoupSessionFeatureInterface *feature_inter
 				      gpointer interface_data)
 {
 	feature_interface->request_queued = soup_cookie_jar_request_queued;
-	feature_interface->request_started = soup_cookie_jar_request_started;
 	feature_interface->request_unqueued = soup_cookie_jar_request_unqueued;
 }
 
@@ -751,16 +751,13 @@ soup_cookie_jar_delete_cookie (SoupCookieJar *jar,
 {
 	SoupCookieJarPrivate *priv;
 	GSList *cookies, *p;
-	char *domain;
 
 	g_return_if_fail (SOUP_IS_COOKIE_JAR (jar));
 	g_return_if_fail (cookie != NULL);
 
 	priv = SOUP_COOKIE_JAR_GET_PRIVATE (jar);
 
-	domain = g_strdup (cookie->domain);
-
-	cookies = g_hash_table_lookup (priv->domains, domain);
+	cookies = g_hash_table_lookup (priv->domains, cookie->domain);
 	if (cookies == NULL)
 		return;
 
@@ -769,7 +766,7 @@ soup_cookie_jar_delete_cookie (SoupCookieJar *jar,
 		if (soup_cookie_equal (cookie, c)) {
 			cookies = g_slist_delete_link (cookies, p);
 			g_hash_table_insert (priv->domains,
-					     domain,
+					     g_strdup (cookie->domain),
 					     cookies);
 			soup_cookie_jar_changed (jar, c, NULL);
 			soup_cookie_free (c);
