@@ -4,6 +4,7 @@
  *
  * Copyright (C) 2001-2004 Novell, Inc.
  * Copyright (C) 2008 Red Hat, Inc.
+ * Copyright (C) 2013 Igalia, S.L.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -14,11 +15,9 @@
 #include <string.h>
 
 #include "soup-logger.h"
-#include "soup-message.h"
-#include "soup-session.h"
-#include "soup-session-feature.h"
-#include "soup-socket.h"
-#include "soup-uri.h"
+#include "soup-connection.h"
+#include "soup-message-private.h"
+#include "soup.h"
 
 /**
  * SECTION:soup-logger
@@ -56,7 +55,7 @@
  * </screen></informalexample>
  *
  * The <literal>Soup-Debug-Timestamp</literal> line gives the time (as
- * a #time_t) when the request was sent, or the response fully
+ * a <type>time_t</type>) when the request was sent, or the response fully
  * received.
  *
  * The <literal>Soup-Debug</literal> line gives further debugging
@@ -70,26 +69,28 @@
  *
  * Currently, the request half of the message is logged just before
  * the first byte of the request gets written to the network (from the
- * #SoupSession::request_started signal), which means that if you have
+ * #SoupMessage::starting signal), which means that if you have
  * not made the complete request body available at that point, it will
- * not be logged. The response is logged just after the last byte of
- * the response body is read from the network (from the
- * #SoupMessage::got_body or #SoupMessage::got_informational signal),
- * which means that the #SoupMessage::got_headers signal, and anything
- * triggered off it (such as #SoupSession::authenticate) will be
- * emitted <emphasis>before</emphasis> the response headers are
- * actually logged.
+ * not be logged.
+ *
+ * The response is logged just after the last byte of the response
+ * body is read from the network (from the #SoupMessage::got_body or
+ * #SoupMessage::got_informational signal), which means that the
+ * #SoupMessage::got_headers signal, and anything triggered off it
+ * (such as #SoupSession::authenticate) will be emitted
+ * <emphasis>before</emphasis> the response headers are actually
+ * logged.
+ *
+ * If the response doesn't happen to trigger the
+ * #SoupMessage::got_body nor #SoupMessage::got_informational signals
+ * due to, for example, a cancellation before receiving the last byte
+ * of the response body, the response will still be logged on the
+ * event of the #SoupMessage::finished signal.
  **/
 
+static SoupSessionFeatureInterface *soup_logger_default_feature_interface;
 static void soup_logger_session_feature_init (SoupSessionFeatureInterface *feature_interface, gpointer interface_data);
 
-static void request_queued  (SoupSessionFeature *feature, SoupSession *session,
-			     SoupMessage *msg);
-static void request_started  (SoupSessionFeature *feature, SoupSession *session,
-			      SoupMessage *msg, SoupSocket *socket);
-static void request_unqueued  (SoupSessionFeature *feature,
-			       SoupSession *session, SoupMessage *msg);
-
 G_DEFINE_TYPE_WITH_CODE (SoupLogger, soup_logger, G_TYPE_OBJECT,
 			 G_IMPLEMENT_INTERFACE (SOUP_TYPE_SESSION_FEATURE,
 						soup_logger_session_feature_init))
@@ -103,6 +104,7 @@ typedef struct {
 	GQuark              tag;
 	GHashTable         *ids;
 
+	SoupSession        *session;
 	SoupLoggerLogLevel  level;
 	int                 max_body_size;
 
@@ -131,7 +133,7 @@ soup_logger_init (SoupLogger *logger)
 }
 
 static void
-finalize (GObject *object)
+soup_logger_finalize (GObject *object)
 {
 	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (object);
 
@@ -156,16 +158,7 @@ soup_logger_class_init (SoupLoggerClass *logger_class)
 
 	g_type_class_add_private (logger_class, sizeof (SoupLoggerPrivate));
 
-	object_class->finalize = finalize;
-}
-
-static void
-soup_logger_session_feature_init (SoupSessionFeatureInterface *feature_interface,
-				  gpointer interface_data)
-{
-	feature_interface->request_queued = request_queued;
-	feature_interface->request_started = request_started;
-	feature_interface->request_unqueued = request_unqueued;
+	object_class->finalize = soup_logger_finalize;
 }
 
 /**
@@ -385,6 +378,9 @@ soup_logger_detach (SoupLogger  *logger,
 	soup_session_remove_feature (session, SOUP_SESSION_FEATURE (logger));
 }
 
+static void soup_logger_print (SoupLogger *logger, SoupLoggerLogLevel level,
+			       char direction, const char *format, ...) G_GNUC_PRINTF (4, 5);
+
 static void
 soup_logger_print (SoupLogger *logger, SoupLoggerLogLevel level,
 		   char direction, const char *format, ...)
@@ -445,19 +441,19 @@ soup_logger_print_basic_auth (SoupLogger *logger, const char *value)
 			*p = '*';
 	}
 	soup_logger_print (logger, SOUP_LOGGER_LOG_HEADERS, '>',
-			   "Authorization: Basic [%.*s]", len, decoded);
+			   "Authorization: Basic [%.*s]", (int)len, decoded);
 	g_free (decoded);
 }
 
 static void
 print_request (SoupLogger *logger, SoupMessage *msg,
-	       SoupSession *session, SoupSocket *socket,
-	       gboolean restarted)
+	       SoupSocket *socket, gboolean restarted)
 {
 	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
 	SoupLoggerLogLevel log_level;
 	SoupMessageHeadersIter iter;
 	const char *name, *value;
+	char *socket_dbg;
 	SoupURI *uri;
 
 	if (priv->request_filter) {
@@ -487,24 +483,43 @@ print_request (SoupLogger *logger, SoupMessage *msg,
 	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, '>',
 			   "Soup-Debug-Timestamp: %lu",
 			   (unsigned long)time (0));
+
+	socket_dbg = socket ?
+		g_strdup_printf ("%s %u (%p)",
+				 g_type_name_from_instance ((GTypeInstance *)socket),
+				 soup_logger_get_id (logger, socket), socket)
+		: NULL;
+
 	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, '>',
-			   "Soup-Debug: %s %u (%p), %s %u (%p), %s %u (%p)%s",
-			   g_type_name_from_instance ((GTypeInstance *)session),
-			   soup_logger_get_id (logger, session), session,
+			   "Soup-Debug: %s %u (%p), %s %u (%p), %s%s",
+			   g_type_name_from_instance ((GTypeInstance *)priv->session),
+			   soup_logger_get_id (logger, priv->session), priv->session,
 			   g_type_name_from_instance ((GTypeInstance *)msg),
 			   soup_logger_get_id (logger, msg), msg,
-			   g_type_name_from_instance ((GTypeInstance *)socket),
-			   soup_logger_get_id (logger, socket), socket,
+			   socket_dbg ? socket_dbg : "cached",
 			   restarted ? ", restarted" : "");
+	g_free (socket_dbg);
 
 	if (log_level == SOUP_LOGGER_LOG_MINIMAL)
 		return;
 
 	if (!soup_message_headers_get_one (msg->request_headers, "Host")) {
+		char *uri_host;
+
+		if (strchr (uri->host, ':'))
+			uri_host = g_strdup_printf ("[%s]", uri->host);
+		else if (g_hostname_is_non_ascii (uri->host))
+			uri_host = g_hostname_to_ascii (uri->host);
+		else
+			uri_host = uri->host;
+
 		soup_logger_print (logger, SOUP_LOGGER_LOG_HEADERS, '>',
-				   "Host: %s%c%u", uri->host,
+				   "Host: %s%c%u", uri_host,
 				   soup_uri_uses_default_port (uri) ? '\0' : ':',
 				   uri->port);
+
+		if (uri_host != uri->host)
+			g_free (uri_host);
 	}
 	soup_message_headers_iter_init (&iter, msg->request_headers);
 	while (soup_message_headers_iter_next (&iter, &name, &value)) {
@@ -581,6 +596,20 @@ print_response (SoupLogger *logger, SoupMessage *msg)
 	}
 }
 
+static void
+finished (SoupMessage *msg, gpointer user_data)
+{
+	SoupLogger *logger = user_data;
+	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+
+	g_mutex_lock (&priv->lock);
+
+	print_response (logger, msg);
+	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "");
+
+	g_mutex_unlock (&priv->lock);
+}
+
 static void
 got_informational (SoupMessage *msg, gpointer user_data)
 {
@@ -589,6 +618,7 @@ got_informational (SoupMessage *msg, gpointer user_data)
 
 	g_mutex_lock (&priv->lock);
 
+	g_signal_handlers_disconnect_by_func (msg, finished, logger);
 	print_response (logger, msg);
 	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "");
 
@@ -623,6 +653,7 @@ got_body (SoupMessage *msg, gpointer user_data)
 
 	g_mutex_lock (&priv->lock);
 
+	g_signal_handlers_disconnect_by_func (msg, finished, logger);
 	print_response (logger, msg);
 	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "");
 
@@ -630,30 +661,14 @@ got_body (SoupMessage *msg, gpointer user_data)
 }
 
 static void
-request_queued (SoupSessionFeature *logger, SoupSession *session,
-		SoupMessage *msg)
+starting (SoupMessage *msg, gpointer user_data)
 {
-	g_return_if_fail (SOUP_IS_MESSAGE (msg));
-
-	g_signal_connect (msg, "got-informational",
-			  G_CALLBACK (got_informational),
-			  logger);
-	g_signal_connect (msg, "got-body",
-			  G_CALLBACK (got_body),
-			  logger);
-}
-
-static void
-request_started (SoupSessionFeature *feature, SoupSession *session,
-		 SoupMessage *msg, SoupSocket *socket)
-{
-	SoupLogger *logger = SOUP_LOGGER (feature);
+	SoupLogger *logger = SOUP_LOGGER (user_data);
+	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
 	gboolean restarted;
 	guint msg_id;
-
-	g_return_if_fail (SOUP_IS_SESSION (session));
-	g_return_if_fail (SOUP_IS_MESSAGE (msg));
-	g_return_if_fail (SOUP_IS_SOCKET (socket));
+	SoupConnection *conn;
+	SoupSocket *socket;
 
 	msg_id = soup_logger_get_id (logger, msg);
 	if (msg_id)
@@ -663,22 +678,69 @@ request_started (SoupSessionFeature *feature, SoupSession *session,
 		restarted = FALSE;
 	}
 
-	if (!soup_logger_get_id (logger, session))
-		soup_logger_set_id (logger, session);
+	if (!soup_logger_get_id (logger, priv->session))
+		soup_logger_set_id (logger, priv->session);
 
-	if (!soup_logger_get_id (logger, socket))
+	conn = soup_message_get_connection (msg);
+	socket = conn ? soup_connection_get_socket (conn) : NULL;
+	if (socket && !soup_logger_get_id (logger, socket))
 		soup_logger_set_id (logger, socket);
 
-	print_request (logger, msg, session, socket, restarted);
+	print_request (logger, msg, socket, restarted);
 	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "");
 }
 
 static void
-request_unqueued (SoupSessionFeature *logger, SoupSession *session,
-		  SoupMessage *msg)
+soup_logger_request_queued (SoupSessionFeature *logger,
+			    SoupSession *session,
+			    SoupMessage *msg)
 {
 	g_return_if_fail (SOUP_IS_MESSAGE (msg));
 
+	g_signal_connect (msg, "starting",
+			  G_CALLBACK (starting),
+			  logger);
+	g_signal_connect (msg, "got-informational",
+			  G_CALLBACK (got_informational),
+			  logger);
+	g_signal_connect (msg, "got-body",
+			  G_CALLBACK (got_body),
+			  logger);
+	g_signal_connect (msg, "finished",
+			  G_CALLBACK (finished),
+			  logger);
+}
+
+static void
+soup_logger_request_unqueued (SoupSessionFeature *logger,
+			      SoupSession *session,
+			      SoupMessage *msg)
+{
+	g_return_if_fail (SOUP_IS_MESSAGE (msg));
+
+	g_signal_handlers_disconnect_by_func (msg, starting, logger);
 	g_signal_handlers_disconnect_by_func (msg, got_informational, logger);
 	g_signal_handlers_disconnect_by_func (msg, got_body, logger);
+	g_signal_handlers_disconnect_by_func (msg, finished, logger);
+}
+
+static void
+soup_logger_feature_attach (SoupSessionFeature *feature,
+			    SoupSession *session)
+{
+	SOUP_LOGGER_GET_PRIVATE (feature)->session = session;
+
+	soup_logger_default_feature_interface->attach (feature, session);
+}
+
+static void
+soup_logger_session_feature_init (SoupSessionFeatureInterface *feature_interface,
+				  gpointer interface_data)
+{
+	soup_logger_default_feature_interface =
+		g_type_default_interface_peek (SOUP_TYPE_SESSION_FEATURE);
+
+	feature_interface->attach = soup_logger_feature_attach;
+	feature_interface->request_queued = soup_logger_request_queued;
+	feature_interface->request_unqueued = soup_logger_request_unqueued;
 }
